package com.holub.login.server;

// (c) 2010 Allen I Holub. All rights reserved.

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import javax.servlet.http.*;
import javax.servlet.ServletContext;

import java.io.*;
import java.net.InetAddress;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.net.UnknownHostException;
import java.security.SecureRandom;
import java.util.Date;

/** This servlet handles server-side communication with the login application. 
 *  See the
 *  <a href="../client/package-summary.html">package overview</a>
 *  for the <code>com.holub.client</code> package for details on how the Login manger is integrated into
 *  the application.
 * <p>
 * The LoginManger is a Gang-of-Four mediator that hides all of the actual
 * communication with the client-side javascript. To use it, you need to
 * plug in an object that implements the
 * <a href="../client.LoginManager.ServerSideLoginSupport.html>
 * <code>com.holub.login.client.LoginManager.ServerSideLoginSupport</code></a> interface.
 * The LoginManger makes calls to that plug in to do the database-related work
 * associated with the login process.
 * <p>
 * This class is an abstract class that you must complete by providing version of several
 * Template Methods that handle database operations and the like. These methods are:
 * <blockquote>
 * {@link #checkPasswordValid(String, String, String)}<br>
 * {@link #verifyLoginID(String, String, String)}<br>
 * {@link #logout(String, String)}<br>
 * {@link #recover(String, String)}<br>
 * {@link #initializeLoginServices(HttpServletRequest)}<br>
 * <blockquote>
 *  All of these methods may throw exceptions, which are just caught and logged by the LoginManger (with
 *  an appropriate error status returned to the actual login client).
 *  <p>
 *  Several support methods return or use a "login-session ID." This ID is
 *  an arbitrary String that uniquely identifies the current login session.
 *  That is, every user should have a unique login-session ID, and that token should
 *  change every time the user logs in. The mechanism is similar to a session id.
 *  <p>
 *  The login-token string is effectively a Gang-of-Four Memento. 
 *  It is stored in a cookie by the login system (a GoF "caretaker") and passed back to the server as needed.
 *  The token doesn't have to be an opaque handle, however.
 *  Since it's passed to the client side of the program as a method argument, that client-side code
 *  can use the information found in it. (The client side code should <em>never</em> modify the token, however.)
 *  Consequently, the login token can contain information of use to the client in addition to a session identifier.
 *  For example, a login token could be a JSON string that tells the client side the user id or user name:
 *  <pre>
 *  {"sessionId":"10-9874514098a-s8-09898709870", "userId":"418", "userName":"fred"}
 *  </pre>
 *  The client-side code can then parse the token string as required. If the user chooses
 *  to remain logged in, the entire token is passed back to the server, so the server
 *  can use the token to store information that facilitates the re-login process.
 *  <p>
 *  See both the 
 *  <a href="../client.LoginPanelConnector.ClientSideLoginSupport.html">
 *  	<code>com.holub.login.client.LoginPanelConnector.ClientSideLoginSupport</code>
 *  </a> 
 *  interface and the
 *  <a href="../client/package-summary.html">package overview</a>
 *  for the <code>com.holub.client</code> package for more details on how the client and server sides of the package interoperate.
 *  <p>
 *  The server should store the token in the database every
 *  time it generates a new token. When a re-login is attempted via a token, the server
 *  should then compare the token passed to {@link #verifyTokenAndGetReplacement(String, String)}
 *  with the version in the database to make sure it's valid. When the user logs out, the
 *  associated token in the database should be nulled out.
 *  <p>
 *  Note that it's not an error for a token to be in the database when the user logs in with a username and password.
 *  What that means is that the user has logged in from another browser instance without having logged out of the first instance.
 *  Unless you've written your application to handle this situation (which is hard to do), the user should effectively
 *  be logged of of the first instance when they log on to the second one. To do that,
 *  generate a new token and return the new token from you login handler. The client application <em>must</em>
 *  pass the login token back to the server with every RPC call to prevent a hacker from making RPC calls without having
 *  logged in.  Similarly, the server should check that the login token is valid for a given user with every RPC call.
 *  For that reason, it's convenient to encode the userID into a JSON login token, as I described earlier.
 *  The RPC call should throw an exception if it encounters a bad login token, and the RPC handler on the client
 *  could just redraw the page to effect a logout.
 * 
 * @author Allen Holub
 */

public abstract class LoginManager extends HttpServlet
{
	public static final long serialVersionUID = 1;
	
	private static final Logger log = LoggerFactory.getLogger(LoginManager.class);
	
	//----------------------------------------------------------------------
	
	/** All requests are made via GET. Log an error if we get a POST */
	public void doPost(HttpServletRequest request, HttpServletResponse response) throws IOException
	{
    	log.error( "Login error: unexpected POST from " + request.getRemoteHost() + ": URL=" +
                                    		request.getRequestURL() + "?" + request.getQueryString() );
	}
	
	public void init()
	{
		ServletContext context = getServletContext();
		System.setProperty("domain.docroot", context.getRealPath("/") ); // used by other servlets in my app. Feel free to delete.
	}
	
	public void destroy(){}
	
	/** Handle the post from the login box. The following fields are expected to exist in the post data:
	 *  <table border="1" >
	 *  
	 *  In the post data, the state="xxx" field specifies the state that the login system is in.
	 *  <table> 
	 *  <tr><td>VALUE	 </td><td>DESCRIPTION					</td><td>OTHER FIELDS USED	</td></tr>
	 *  <tr><td>login	 </td><td>A user login using password	</td><td>userName, password	</td></tr>
	 *  <tr><td>verify	 </td><td>A user login using session ID	</td><td>userName, sessionID</td></tr>
	 *  <tr><td>logout	 </td><td>logout						</td><td>sessionID			</td></tr>
	 *  <tr><td>recover	 </td><td>recover forgotten password 	</td><td>userName			</td></tr>
	 *  <tr><td></table> 
	 *  <p>
	 *  The methods "return" values by writing to the output stream. The first line of the output
	 *  will hold one of the strings:
	 *  "OK"		Everything worked correctly. Subsequent lines hold the response.
	 *  "FAIL"	 	All subsequent lines comprise an error message suitable for display to the user.
	 *  "ERROR"		probably an exception toss. The exception is logged, but nothing is returned
	 *  			to the user.
	 *  <p>
	 *  Logout always returns OK, and the associated code clears out the loginToken. A FAIL
	 *  response from a "forgot" request means that the user name was unrecognized.
	 */
	public void doGet(HttpServletRequest request, HttpServletResponse response) throws IOException
	{
    	PrintWriter out 		= response.getWriter();
    	String 		output		= null;
    	
		try
		{
			System.setProperty( "domain.name", request.getRequestURL().toString() );
		
	    	String action			= request.getParameter("action");
	    	String userName			= request.getParameter("username");
	    	String password			= request.getParameter("password");
	    	String loginSessionID 	= request.getParameter("sessionID");
	    	
	    	if( loginSessionID != null )
		    	loginSessionID = unescape( loginSessionID );
	    	
	        log.debug(	  "action="			+	(action 		==null ? "<null>" : ("[" + action   + "]")) +
	         			", userName="		+	(userName		==null ? "<null>" : ("[" + userName + "]")) +
	         			", password="		+	(password		==null ? "<null>" : ("[" + password + "]")) +
	         			", loginSessionID="	+	(loginSessionID ==null ? "<null>" : ("[" + loginSessionID + "]")) );
	        
	        if( action == null )
	        {	
	        	log.error("No action requested; returning FAIL status.");
	        	output =  "{ 'status':'FAIL' }";
	        }
	        else
	        {
	    		checkNotEmpty( "userName", userName );
	    		
	    		initializeLoginServices( request );
	    		
		        if( action.equals("login") )
		    	{
		    		checkNotEmpty( "password", password );
		    		output = doLogin( userName, password );
		    	}
		        else if( action.equals("verify") )
		        {
		    		checkNotEmpty("loginSessionID", loginSessionID);
		    		output = doVerify( userName, loginSessionID );
		        }
		    	else if( action.equals("logout") )
		    	{	
		    		checkNotEmpty("loginSessionID", loginSessionID);
		    		output = doLogout( userName, loginSessionID ); 
		    	}
		    	else if( action.equals("recover") )
		    	{	
		    		output = doRecover( userName, request.getRequestURL().toString() );  // request.getRequestURI() );
		    	}
		    	else
		    	{	log.error("Unrecognized request to login servlet: " + action );
		    	}
	        }
		}
		catch( Exception e )
		{
			output = "{ 'status':'FAIL' }";
			log.error("Exception from LoginManager servlet:\n" + stackTrace(e) );
		}
        
        response.setContentType( "text/html; charset=us-ascii" );
		out.println( output );
		out.close();
	}
	
	/** Verify that the variable isn't empty (is not null and is not zero length).
	 *  Throw an Exception if it's empty.
	 *  
	 * @param what	The variable name. Used in the Exception's message.
	 * @param variable The contents of that variable.
	 * @throws Exception if the contents are null or empty.
	 */
	private final void checkNotEmpty( String what, String variable) throws Exception
	{
		if( variable == null || variable.trim().length() == 0 )
			throw new Exception("Unexpected null or empty "+ what +" in logout request");
	}
	    		
	//----------------------------------------------------------------------
	private final String doLogin( String userName, String password ) throws Exception
	{
		String newLoginSessionID = getSessionID();
		String userData		 	 = checkPasswordValid(userName, password, newLoginSessionID );
		
		log.info( "Login by password: " + userName + "/" + password + ": " + (userData==null?"FAIL":"OK") );
		
		return ( userData == null )
				? "{ 'status':'FAIL','userData':'' }"
				: "{ 'status':'OK'," 	+
				    "'userData':'"    	+ escape(userData) +"'," +
				    "'loginSessionID':'"+ newLoginSessionID +"'" +
		    	 " }"
				;
	}
	
	private final String doVerify(String userName, String oldLoginSessionID ) throws Exception
	{
		String newLoginSessionID = getSessionID();
		String userData 		 = verifyLoginID( userName, oldLoginSessionID, newLoginSessionID );
		
		log.info( "Login by sessionID: "+ userName + "/" + oldLoginSessionID + ": " + (userData==null?"FAIL":"OK") + ". Returning " + userData );
		
		return ( userData == null )
				? "{ 'status':'FAIL','userData':'' }"
				: "{ 'status':'OK'," 	+
				    "'userData':'" 		+ escape(userData) +"'," +
				    "'loginSessionID':'"+ newLoginSessionID +"'" +
		    	 " }"
				;
	}

	/** Perform a logout.
	 */
	private final String doLogout( String userName, String loginSessionID ) throws Exception
	{
		boolean isSuccessful = logout( userName, loginSessionID );
		
		log.info( "Processed logout request from " + userName + "/" + loginSessionID + ": " + (!isSuccessful?"FAIL":"OK") );
		return ( isSuccessful )
				? "{ 'status':'OK',  }"
				: "{ 'status':'FAIL' }"
				;
	}
	
	private final String doRecover( String userName, String requestURI ) throws Exception
	{
		String message = recover(userName, requestURI);
		
		log.info("Recovering password for " + userName + ". Response message=" + message );
		
		return ( message == null )
				? "{ 'status':'FAIL', 'message':'Couldn't recover password for "+userName+" ' }"
				: "{ 'status':'OK',   'message':'"+ escape(message) +"' }"
				;
	}
	//----------------------------------------------------------------------
	/** Return a secure "hash" of a password. This string is, effectively, a
	 *  random string of characters that serves a kind of "signature," uniquely
	 *  identifying the password.
	 *  When you create an account, you should pass the password supplied
	 *  by the user to the current method, and then store the "hash" that this
	 *  method returns in the database. Do not store the plaintext password in the
	 *  database.
	 *  
	 * @param plainTextPassword	A user-supplied password.
	 * @see #verifyHash(String, String)
	 * @return A "secure hash" of the password
	 */
	
	public static String hashPassword( String plainTextPassword )
	{
		return BCrypt.hashpw( plainTextPassword, BCrypt.gensalt());
	}
	
	/** Verify that the indicated plainTextPassword matches the hash
	 *  (which is fetched form the database)
	 *  @see #hashPassword(String)
	 *  
	 *  @param plainTextPassword A password supplied by the user during the login
	 *  	process. When the account was set up, this same password was passed through
	 *  	{@link #hashPassword(String)} to generate a hash, and that hash was stored
	 *  	in the database instead of the plaintext password.
	 *  	Do not store the plaintext password in the database.
	 *  
	 *  @param hash The hash returned from a previous {@link #hashPassword(String)} call
	 *  	when it was passed the password in <code>plaintextPassword</code>.
	 *  
	 *  @return true if the hash matches the original password.
	 */
	
	public static boolean verifyHash( String plainTextPassword, String hash )
	{	return BCrypt.checkpw( plainTextPassword, hash );
	}
	
	//----------------------------------------------------------------------
	/** A utility method that generates a universally-unique
	 *  ID, suitable for use as a session key.
	 *  This method should never return the same value twice.
	 *  The returned ID is not just a random number (which can repeat),
	 *  but is rather a combination of:
	 *  <ul>
	 *  <li>The time/date when the method is called.
	 *  <li>A counter value, incremented on each call. Since the time has
	 *      millisecond-resolution, the counter handles the
	 *      multiple-calls-within-a-millisecond problem.
	 *  <li>The IP address of the server (IPV6 is handled just fine), which
	 *  	handles the two-calls-at-the-same-time-on-different-servers
	 *  	problem.
	 *  <li>A sequence of randomly generated byte values, which should
	 *  	take care of everything else that might come up.
	 *  </ul>
	 *  This value is used as the login-session ID, but feel free to use
	 *  it for other situations. Since the ID is unique across all the
	 *  machines that comprise your computing cluster, you can use
	 *  it as a database or memcache primary key.
	 * 
	 *  @param keyLen The length of the returned String. The minimum length is 48 characters,
	 *  which will have a two-byte random component in the worst case (an IPV6 server address).
	 *  Longer is better. If the <code>keyLen</code> &le; zero, a default length
	 *  ({@link #DEFAULT_SESSION_ID_LENGTH}, currently {@value #DEFAULT_SESSION_ID_LENGTH} characters)
	 *  is used.
	 *  
	 *  @throws illegalArgumentException if 0&gt;<code>keyLen</code>&lt;48
	 *  
	 *  @return The session id.
	 */
	
	public static String getSessionID( int keyLen )
	{	
		if( keyLen <= 0 )
			keyLen = DEFAULT_SESSION_ID_LENGTH;
		
		if( keyLen < 48  )
			throw new IllegalArgumentException("Illegal ID length. Must be >= 48");
				
		// Create a suffix from the server IP address, time, and a counter. The time is truncated to 32 bits, but that's
		// adequate for our purposes.
		//
		StringBuffer suffix = new StringBuffer();
		hexEncode	 ( (int)(new Date().getTime()),	4, suffix ); // 8 characters 						(4 bytes)
		hexEncode	 ( (short)counter++,			2, suffix ); // 4 characters						(2 bytes)
		suffix.append( serverIPAddress 						  ); // 8 (IPV4) or 32 (IPV6) characters.	(4 or 16 bytes)
			
		byte[] randomValue = new byte[ (keyLen - suffix.length()) / 2 ]; // need half as many bytes as characters (each byte is encoded as two characters).
		random.nextBytes ( randomValue );
		return byteEncode( randomValue ) + suffix.toString(); 	// Putting the random value first improves hash-table-lookup time. Base-64 encoding
																// would be more compact than byte encoding, but it's not worth the trouble.
	}
	
	public static final int DEFAULT_SESSION_ID_LENGTH = 56;
	
	/** Convenience method. Same as {@link #getSessionID(int) getSessionID(DEFAULT_SESSION_ID_LENGTH)}.
	 */
	
	public static String getSessionID(){ return getSessionID(DEFAULT_SESSION_ID_LENGTH); }
	
	/** 
	 * Alias for {@link #getSessionID()}. User-supplied suffixes are no longer supported.
	 * @deprecated in favor of {@link #getSessionID()}.
	 */
	public static String getSessionID(String suffix)
	{	return getSessionID();
	}
	
	/** Encode an array of bytes into its hex representation (each byte is 2 hex digits).
	 */
	private static String byteEncode( byte[] input )
	{
		StringBuffer output = new StringBuffer();
		for( byte c : input )
			hexEncode( c, 1, output );
		return output.toString();
	}
	
	/** Encode the specified number of bytes in a 32-bit (4-byte) int into hex
	 *  @param value The value to encode
	 *  @param numberOfBytes The number of (least significant) bytes to encode 
	 *  @param target place to put the result.
	 */
	
	private static void hexEncode( int value, int numberOfBytes, StringBuffer target )
	{
		while( --numberOfBytes >= 0 )
		{
			target.append( hexAlphabet[ (value >> 4) & 0x0f ] );
			target.append( hexAlphabet[ (value     ) & 0x0f ] );
			value >>>= 8 ;
		}
	}
	
	/** Length of the String returned from {@link #getSessionID()} */
	private static		 short			counter;						// 16 bits (4 characters when encoded)
	private static final SecureRandom	random = new SecureRandom();
	private static final char[]			hexAlphabet = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c', 'd', 'e', 'f' };
	private static final String 		serverIPAddress;
	static
	{	
		String hostAddr = "";
		try
		{	hostAddr = byteEncode( InetAddress.getLocalHost().getAddress() );
		}
		catch( UnknownHostException e )
		{	log.error("Can't get host address!, Session ID's may repeat acrosss multiple machines.");
		}
		serverIPAddress = hostAddr;
	}
	
	//----------------------------------------------------------------------
	/** Simulate a javaScript escape() call. This is identical to url encoding, except that
	 *  a space is encoded as %20 instead of +.
	 */
	public static String escape( String s )
	{
		try
		{
			String encoded = URLEncoder.encode( s, "UTF-8" );	// do everything except the spaces
			encoded = encoded.replace( "+", "%20" );			// do the spaces
			return encoded;
		}
		catch( UnsupportedEncodingException e )
		{	log.error("Internal error: UTF-8 not supported by URLEncoder");
		}
		return null;
	}
	
	/** Reverse whatever {@link #escape(String)} did.
	 */
	public static String unescape( String s )
	{
		try
		{
			s = s.replace( "%20", "+" );
			return URLDecoder.decode( s, "UTF-8" );
		}
		catch( UnsupportedEncodingException e )
		{	log.error("Internal error: UTF-8 not supported by URLEncoder");
		}
		return null;
	}
	
	//----------------------------------------------------------------------
	/** Return a string that holds an exception message and the associated
	 *  stack trace so that we can log it.
	 */
	
	public static String stackTrace( Throwable e )
	{
		StringWriter s = new StringWriter();
		PrintWriter  p = new PrintWriter(s);
		p.println( e.getMessage() );
		e.printStackTrace(p);
		return s.toString();
	}
	
	//----------------------------------------------------------------------
	// Methods that must be provided by the derived class.
	
	/**
	 * Performs the initial user/password validation.
	 * If the password is valid, return a non-null String which will be passed
	 * back to the client (for use on the main page). Note that the page to which this
	 * String is passed might not be secure (loaded using HTTPS), so you shouldn't
	 * return anything in this string that you need to keep hidden.
	 * <p>
	 * A user's actual password should never be stored in the database. Instead,
	 * you should hash the password using {@link LoginManager#hashPassword(String)})
	 * and then store the hash. To log in, compare the password argument
	 * of this method to the hashed value with {@link #verifyHash(String, String)}.
	 * 
	 * @param userName  The user name supplied by the end user.
	 * @param password  The plain-text (not hashed) password supplied by the end user.
	 * @param sessionID This login-session ID will be stored on the client, and you should
	 * 					also store it in the database. If the user has chosen to remain logged in,
	 * 					this sessionID is passed back to user via a {@link #verifyLoginID(String, String)} call.
	 * @return	An arbitrary String (could be JSON) that's passed to the login client,
	 * 			or <code>null</code> if the login attempt fails.
	 * 			This string might be passed to the
	 * 			insecure part of an AJAX page, so you should not put anything sensitive
	 * 			(e.g. the sessionID) into it.
	 */
	public abstract String checkPasswordValid(String userName, String password, String loginSessionID) throws Exception;
	
	/** Checks whether a login-session ID (stored on the client) is still valid. This method
	 *  is called from the login subsystem when a user has checked the "remain logged in"
	 *  box, left the site, and then returned to the site at a later date. Note that
	 *  all users effectively remain logged in for about five minutes after leaving
	 *  a site so that they can return from a link without having to log in again.
	 *  However, if they log out, they will be logged out (the client-side login-session ID
	 *  will be cleared, in this case). This method should compare the login-session ID
	 *  with a login-session ID that was stored in the database when the user logged in with
	 *  his/her password.
	 *  <p>
	 *  The calling method in the base class will replace the <code>oldLoginSessionID</code> with a
	 *  new one, which will be used on a subsequent login. Consequently, you should update 
	 *  the "current login-session ID" in the database to hold the <code>newLoginSessionID</code>.
	 * 
	 * @param userName The user to whom the loginSessionID was issued.
	 * @parem oldLoginSessionID This is the login-session ID that's currently stored on the client (and also, in
	 * 		  the database). It was passed to {@link #checkPasswordValid(String, String, String)} when
	 * 		  the initial login occurred.
	 * @parem newLoginSessionID This is the login-session ID that will replace <code>oldLoginSessionID</code>
	 * 		  on the client when the re-login succeeds. You should replace the old login-session ID with this one
	 * 		  in the database if the login is successful.
	 * @return  Identical to the return value from {@link #checkPasswordValid(String, String, String)}.
	 * 
	 * @see com.holub.login.server.LoginManager#getSessionID(String)
	 */
	public abstract String verifyLoginID(String userName, String oldLoginSessionID, String newLoginSessionID ) throws Exception;
	
	/** Log out the user.
	 * 
	 * @param userName
	 * @param loginSessionID  The login-session ID passed to the most recent call to
	 * 			{@link #checkPasswordValid(String, String, String)} or
	 * 			{@link #verifyLoginID(String, String, String)}.
	 * @return true if the logout was successful. Return false otherwise.
	 */
	public abstract boolean logout(String userName, String loginSessionID ) throws Exception;
	
	/** 
	 *  Do something to recover the password, and return a String that can be used
	 *  on the client side. Typically, that string will just be a block of HTML that the
	 *  client side displays in a dialog box, but it could also be something like a JSON
	 *  string that the client side parses.
	 *  That string could specify an entire password-recover dialog in HTML if more
	 *  work has to be done.
	 *  The <code>String</code> is passed to the client side 
	 *  <a href="../client/LoginPanelConnector.ClientSideLoginSupport.html#recoverPassword(java.lang.String)">
	 *  <code>LoginPanelConnector.ClientSideLoginSupport.recoverPassword(String)</code></a>
	 *  method but other than that, the login system doesn't do anything with it.
	 *  
	 * @param userName user name supplied by the actual user.
	 * @param requestURL the entire request URL (as returned from HTTPServletResponse.getRequestURL()) for this password-recovery request.
	 * 
	 * @return null if the password couldn't be recovered (if the userName is
	 * 			not valid, for example) or a message string to send to the client side
	 * 			if the recovery process worked of if further work needs doing.
	 */
	
	public abstract String recover(String userName, String requestURL ) throws Exception;
	
	/** Called to do any server-side initialization (such as setting up database connections). This
	 *  method is called exactly once, and is called before any of the other abstract methods defined in this
	 *  class are called.
	 * @param request The request object passed to the servlet the first time a login request is made.
	 */
	public abstract void initializeLoginServices( HttpServletRequest request ) throws Exception;
}
