package com.holub.login.server;

import javax.servlet.http.HttpServletRequest;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/** This class provides a stub for (and an example of) how to hook server-side login support up to the
 *  login-manager servlet. You can register this class as a servlet to test the system. You can
 *  also extend this class to inherit default test methods if you want to use them. The default
 *  implementations of the methods log a "severe" error (using the java.util.logging system). For
 *  testing purposes, they recognize a username and password of "user" and "password" as valid and
 *  everything else as invalid, and correctly process a re-login if remain-logged-in is specified.
 *  <p>
 *  The login-test methods return a small JSON String of the form:
 *  <pre>
	{'value':'User-supplied login value (login 1)'}
 *  </pre>
 *  where the number is incremented on every login attempt. The recover-password method returns
 *  the following HTML:
 *  <pre>
 *  {@value #PASSWORD_RECOVERY_STRING}
 *  </pre>
 * 
 * @author allen
 */

public class LoginManagerAdapter extends LoginManager
{
	private static final long serialVersionUID = 1L;
	private static final Logger log = LoggerFactory.getLogger(LoginManager.class);

	private static String currentHashedPassword = LoginManager.hashPassword("password");  // STORE THE HASHED VALUE IN THE DATABASE !
	private static String currentLoginSessionID = null ;
	private static int	  count = 0;
	
	private static final String PASSWORD_RECOVERY_STRING = "<b>Internal \"error\"</b>: Could not recover password." ;
	
	//----------------------------------------------------------------------
	// ServerSideLoginSupport overrides
	//----------------------------------------------------------------------
	
	@Override public String checkPasswordValid(String userName, String password, String loginSessionID) throws Exception
	{
		log.error("**** SIMULATING **** login on server: " + userName + "/newID=" + loginSessionID  );
		
		// Verify the plain-text candidate password against the hash stored in the database
		//
		if( !(userName.equals("user") && LoginManager.verifyHash(password, currentHashedPassword)) )
			return null;
		
		// In the real code, this value should be cached in memory and backed up to the database. You'll
		// need it to handle the remain-logged-in feature. Do not return this value in the return-value String.
		//
		currentLoginSessionID = loginSessionID;
		
		return "{'value':'User-supplied login value (login " + ++count + ")'}";
	}

	@Override public String verifyLoginID(String userName, String oldLoginSessionID, String newLoginSessionID) throws Exception
	{
		log.error("**** SIMULATING **** verify on server: " + userName + "/oldID=" + oldLoginSessionID +"/newID=" + newLoginSessionID );
		
		// If the presented (old) session ID matches the cached one, then we can log in.
		//
		if( !oldLoginSessionID.equals(currentLoginSessionID) )
			return null;
		
		// Replace the old session ID with a new one. Not, strictly speaking, necessary, but it improves security.
		//
		currentLoginSessionID = newLoginSessionID;
		return "{'value':'User-supplied login value (verify " + ++count + ")'}";
	}

	@Override public boolean logout(String userName, String loginSessionID )
	{
		// if the session ID doesn't match the cached one, then this request is a hacker trying to log off someone else!
		
		log.error("**** SIMULATING **** logout on server: " + userName + "/" + loginSessionID  );
		return true;
	}

	@Override public String recover(String userName, String requestURI )
	{
		log.error("**** SIMULATING **** password recovery on server ("+ requestURI +"). user=" + userName + ". Returning: " + PASSWORD_RECOVERY_STRING );
		return PASSWORD_RECOVERY_STRING;
	}

	public void initializeLoginServices(HttpServletRequest request)
	{	log.info("Initializing login-support from class: " + getClass().getName() );
	}
}
